/* copyright (C) 2001 Jean-Luc Fontaine (mailto:jfontain@free.fr) */
/* this library is free software: please read the README file enclosed in this package */

/* $Id: tclpython.c,v 1.20 2002/11/08 09:43:16 jfontain Exp $ */

/*

Provide Python interpreters accessible from Tcl as a package named "tclpython".

Created using the following command:

# Linux Redhat 8.0:
$ cc -shared -s -o tclpython.so.2.1 -fPIC -O2 -Wall -I/usr/include/python2.2 tclpython.c -L/usr/lib/python2.2/config -lpython2.2 -lpthread -lutil

# BSD (example, thanks to Dave Bodenstab):
$ cc -fpic -I/usr/local/include/python -I/usr/local/include/tcltk/tcl8.3 -c tclpython.c
$ ld -o tclpython.so -Bshareable -L/usr/X11R6/lib -L/usr/local/lib -L/usr/local/share/python/config tclpython.o -lpython -lutil -lreadline -ltermcap -lcrypt -lgmp -lgdbm -lpq -lz -ltcl83 -ltk83 -lX11

*/


#include <tcl.h>
/* declare a few functions so that tclInt.h include is not necessary, otherwise Tcl sources would be required to compile this */
extern Tcl_Namespace *Tcl_FindNamespace(Tcl_Interp *, char *, Tcl_Namespace *, int);
extern Tcl_Namespace *Tcl_CreateNamespace (Tcl_Interp *, char *, ClientData, Tcl_NamespaceDeleteProc *);
extern int Tcl_Export(Tcl_Interp *, Tcl_Namespace *, char *, int);
#include <Python.h>
#include <cStringIO.h>

static unsigned existingInterpreters = 0;
static struct Tcl_HashTable threadStates;
static struct Tcl_HashTable dictionaries;
static int newIdentifier;
#ifdef WITH_THREAD
static PyThreadState *globalState = 0;
#endif


static int pythonInterpreter(ClientData clientData, Tcl_Interp *interpreter, int numberOfArguments, Tcl_Obj * CONST arguments[])
{
    char *command;
    int identifier;
    PyObject *output;
    PyObject *message;
    PyObject *result;
    PyObject *globals;
    char *string = 0;
    int length;
    Tcl_Obj *object;
    struct Tcl_HashEntry *entry;
#ifdef WITH_THREAD
    PyThreadState *state;
#endif

    if(numberOfArguments != 3){
        Tcl_WrongNumArgs(interpreter, 1, arguments, "eval script");
        return TCL_ERROR;
    }
    command = Tcl_GetString(arguments[1]);
    if(strcmp(command, "eval") != 0){
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(object, "bad option \"", command, "\": must be eval", 0);
        Tcl_SetObjResult(interpreter, object);
        return TCL_ERROR;
    }
    identifier = atoi(Tcl_GetString(arguments[0]) + 6);                              /* interpreter and command name is "pythonN" */
    entry = Tcl_FindHashEntry(&threadStates, (ClientData)identifier);
    if(entry == 0){
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(
            object, "invalid interpreter \"", Tcl_GetString(arguments[0]), "\": internal error, please report to tclpython author",
            0
        );
        Tcl_SetObjResult(interpreter, object);
        return TCL_ERROR;
    }
    globals = Tcl_GetHashValue(Tcl_FindHashEntry(&dictionaries, (ClientData)identifier));
#ifdef WITH_THREAD
    state = Tcl_GetHashValue(entry);
    PyEval_RestoreThread(state);                              /* acquire the global interpreter lock and make this thread current */
#endif
    result = PyRun_String(Tcl_GetString(arguments[2]), Py_file_input, globals, globals);
    if(result == 0){                                                                                          /* an error occured */
        output = PycStringIO->NewOutput(1024);               /* use a reasonable initial size but big enough to handle most cases */
        PySys_SetObject("stderr", output);                                                /* capture all interpreter error output */
        PyErr_Print();                                            /* so that error is printed on standard error, redirected above */
        message = PycStringIO->cgetvalue(output);
        string = PyString_AsString(message);
        length = PyString_Size(message);
        if((length > 0) && (string[length - 1] == '\n')) length--;               /* eventually remove trailing new line character */
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(object, Tcl_GetString(arguments[0]), ": ", 0);                    /* identify interpreter in error */
        Tcl_AppendObjToObj(object, Tcl_NewStringObj(string, length));
        Py_DECREF(output);
    } else {
        output = PyDict_GetItemString(globals, "_");                          /* get true result from patched Python interpreter: */
        /* in ceval.c, eval_code2():
            case POP_TOP:
                v = POP();
                PyDict_SetItemString(f->f_globals, "_", v);                   // (this line added) store result in _ global variable
                Py_DECREF(v);
                continue;
        */
        if(output == 0)                                                                      /* convert no result to empty output */
            object = Tcl_NewObj();
        else{
            string = PyString_AsString(PyObject_Str(output));
            object = Tcl_NewStringObj(string, -1);
        }
        Py_DECREF(result);
    }
#ifdef WITH_THREAD
    PyEval_SaveThread();                        /* eventually restore the previous thread and release the global interpreter lock */
#endif
    Tcl_SetObjResult(interpreter, object);
    return(result == 0? TCL_ERROR: TCL_OK);
}

static int newInterpreter(Tcl_Interp *interpreter)
{
    int identifier;
    Tcl_Obj *object;
    int created;
#ifdef WITH_THREAD
    PyThreadState *state;
#endif

    identifier = newIdentifier;
#ifndef WITH_THREAD
    if(existingInterpreters > 0){
        Tcl_SetResult(
            interpreter,
            "cannot create several concurrent Python interpreters\n(Python library was compiled without thread support)",
            TCL_STATIC
        );
        return TCL_ERROR;
    }else{
        Py_Initialize();                                                                           /* initialize main interpreter */
        PycString_IMPORT;
    }
    Tcl_SetHashValue(Tcl_CreateHashEntry(&threadStates, (ClientData)identifier, &created), 0);
#else
    if(existingInterpreters == 0){
        Py_Initialize();                                                                           /* initialize main interpreter */
        PyEval_InitThreads();                                               /* initialize and acquire the global interpreter lock */
        PycString_IMPORT;
        globalState = PyThreadState_Swap(0);                                                            /* save the global thread */
    }else{
        PyEval_AcquireLock();                                           /* needed in order to be able to create a new interpreter */
    }
    if(PycStringIO == 0){                                                /* make sure string input/output is properly initialized */
        Tcl_SetResult(interpreter, "fatal error: could not initialize Python string input/output module", TCL_STATIC);
        return TCL_ERROR;
    }
    state = Py_NewInterpreter();          /* hangs here if automatic 'import site' on a new thread is allowed (set Py_NoSiteFlag) */
    if(state == 0){
        PyEval_ReleaseLock();
        Tcl_SetResult(interpreter, "could not create a new interpreter: please report to tclpython author", TCL_STATIC);
        return TCL_ERROR;
    }
    PyEval_ReleaseLock();                                                                  /* release the global interpreter lock */
    Tcl_SetHashValue(Tcl_CreateHashEntry(&threadStates, (ClientData)identifier, &created), state);
#endif
    Tcl_SetHashValue(
        Tcl_CreateHashEntry(&dictionaries, (ClientData)identifier, &created), PyModule_GetDict(PyImport_AddModule("__main__"))
    );
    object = Tcl_NewStringObj("python", -1);
    Tcl_AppendObjToObj(object, Tcl_NewIntObj(identifier));                          /* return "pythonN" as interpreter identifier */
    Tcl_SetObjResult(interpreter, object);
    Tcl_CreateObjCommand(interpreter, Tcl_GetString(object), pythonInterpreter, 0, 0);      /* create command for new interperter */
#ifdef WITH_THREAD
    newIdentifier++;
#endif
    existingInterpreters++;
    return TCL_OK;
}

static int deleteInterpreters(Tcl_Interp *interpreter, int numberOfArguments, Tcl_Obj * CONST arguments[])
{
    int index;
    char *name;
    int identifier;
    struct Tcl_HashEntry *entry;
    Tcl_Obj *object;
#ifdef WITH_THREAD
    PyThreadState *state;
#endif

    for(index = 0; index < numberOfArguments; index++){
        name = Tcl_GetString(arguments[index]);                                                  /* interpreter name is "pythonN" */
        entry = 0;
        if(sscanf(name, "python%u", &identifier) == 1){
            identifier = atoi(name + 6);
            entry = Tcl_FindHashEntry(&threadStates, (ClientData)identifier);
        }
        if(entry == 0){
            object = Tcl_NewObj();
            Tcl_AppendStringsToObj(object, "invalid interpreter \"", name, 0);
            Tcl_SetObjResult(interpreter, object);
            return TCL_ERROR;
        }
#ifdef WITH_THREAD
        state = Tcl_GetHashValue(entry);
        PyEval_AcquireThread(state);                          /* acquire the global interpreter lock and make this thread current */
        Py_EndInterpreter(state);
        PyEval_ReleaseLock();
#endif
        Tcl_DeleteHashEntry(entry);
        Tcl_DeleteHashEntry(Tcl_FindHashEntry(&dictionaries, (ClientData)identifier));
        existingInterpreters--;
        if(existingInterpreters == 0){                                                           /* no remaining sub-interpreters */
#ifdef WITH_THREAD
            PyEval_AcquireThread(globalState);                                                    /* required before finalization */
            globalState = 0;
#endif
            Py_Finalize();                                                                                 /* clean everything up */
        }
    }
    return TCL_OK;
}

static int command(ClientData clientData, Tcl_Interp *interpreter, int numberOfArguments, Tcl_Obj * CONST arguments[])
{
    char *command;
    unsigned new = 0;
    unsigned delete = 0;
    Tcl_Obj *object;

    if(numberOfArguments < 2){
        Tcl_WrongNumArgs(interpreter, 1, arguments, "new|delete ?interp interp ...?");
        return TCL_ERROR;
    }
    command = Tcl_GetString(arguments[1]);
    new = (strcmp(command, "new") == 0);
    delete = (strcmp(command, "delete") == 0);
    if(!new && !delete){
        object = Tcl_NewObj();
        Tcl_AppendStringsToObj(object, "bad option \"", command, "\": must be new or delete", 0);
        Tcl_SetObjResult(interpreter, object);
        return TCL_ERROR;
    }
    if(new){
        if(numberOfArguments != 2){
            Tcl_WrongNumArgs(interpreter, 1, arguments, "new");
            return TCL_ERROR;
        } else {
            return newInterpreter(interpreter);
        }
    }
    if(delete){
#ifdef WITH_THREAD
        if(numberOfArguments < 3){
            Tcl_WrongNumArgs(interpreter, 1, arguments, "delete ?interp interp ...?");
#else
        if(numberOfArguments != 3){                                                          /* there can be one interpreter only */
            Tcl_WrongNumArgs(interpreter, 1, arguments, "delete interp");
#endif
            return TCL_ERROR;
        } else {
            return deleteInterpreters(interpreter, numberOfArguments -2, arguments + 2);
        }
    }
    return TCL_ERROR;                                                                                            /* never reached */
}

int Tclpython_Init(Tcl_Interp *interpreter)
{
    Tcl_Namespace *namespace;
    int code;

    Tcl_InitHashTable(&threadStates, TCL_ONE_WORD_KEYS);
    Tcl_InitHashTable(&dictionaries, TCL_ONE_WORD_KEYS);
    newIdentifier = 0;
    if((namespace = Tcl_FindNamespace(interpreter, "::python", 0, 0)) == 0)         /* if the python namespace does not yet exist */
        namespace = Tcl_CreateNamespace(interpreter, "::python", 0, 0);    /* create it as a direct child of the global namespace */
    if(namespace == 0){
        Tcl_SetObjResult(interpreter, Tcl_NewStringObj("fatal error: could not create python namespace", -1));
        return TCL_ERROR;
    }
    Tcl_CreateObjCommand(interpreter, "::python::interp", command, 0, 0);
    code = Tcl_Export(interpreter, namespace, "interp", 1);                                          /* export the interp command */
    if(code != TCL_OK) return code;
    Py_NoSiteFlag = 1;                      /* suppress automatic 'import site' to prevent interpreter from hanging on new thread */
    return Tcl_PkgProvide(interpreter, "tclpython", "2.1");
}
